@@ -27,6 +27,8 @@
 #include <libglademm.h>
 #include <gconfmm.h>
 #include <libintl.h>
+#include <dbus/dbus-glib.h>
+#include <gdk/gdkx.h>
 
 #define PA_GCONF_ROOT "/system/pulseaudio"
 #define PA_GCONF_PATH_MODULES PA_GCONF_ROOT"/modules"
@@ -36,18 +38,28 @@ public:
     MainWindow(BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade::Xml>& x);
     static MainWindow* create();
 
-    Gtk::EventBox *titleEventBox;
-    Gtk::Button *closeButton;
+    Gtk::Button
+        *closeButton,
+        *zeroconfDiscoverInstallButton,
+        *zeroconfRaopDiscoverInstallButton,
+        *remoteInstallButton,
+        *zeroconfPublishInstallButton,
+        *upnpInstallButton,
+        *rtpRecvInstallButton,
+        *rtpSendInstallButton;
 
     Gtk::CheckButton
         *remoteAccessCheckButton,
         *zeroconfPublishCheckButton,
         *zeroconfDiscoverCheckButton,
+        *zeroconfRaopDiscoverCheckButton,
         *anonymousAuthCheckButton,
         *rtpReceiveCheckButton,
         *rtpSendCheckButton,
         *rtpLoopbackCheckButton,
-        *combineCheckButton;
+        *combineCheckButton,
+        *upnpMediaServerCheckButton,
+        *upnpNullSinkCheckButton;
 
     Gtk::RadioButton
         *rtpMikeRadioButton,
@@ -59,46 +71,79 @@ public:
     bool ignoreChanges;
 
     void onCloseButtonClicked();
+
     void updateSensitive();
+
     void onChangeRemoteAccess();
     void onChangeZeroconfDiscover();
+    void onChangeZeroconfRaopDiscover();
     void onChangeRtpReceive();
     void onChangeRtpSend();
     void onChangeCombine();
+    void onChangeUpnp();
+
+    void onZeroconfDiscoverInstallButtonClicked();
+    void onZeroconfRaopDiscoverInstallButtonClicked();
+    void onRemoteInstallButtonClicked();
+    void onZeroconfPublishInstallButtonClicked();
+    void upnpInstallButtonClicked();
+    void rtpRecvInstallButtonClicked();
+    void rtpSendInstallButtonClicked();
+
     void readFromGConf();
+
     void checkForModules();
+
     void writeToGConfRemoteAccess();
     void writeToGConfZeroconfDiscover();
+    void writeToGConfZeroconfRaopDiscover();
     void writeToGConfRtpReceive();
     void writeToGConfRtpSend();
     void writeToGConfCombine();
+    void writeToGConfUPnP();
+
     void onGConfChange(const Glib::ustring& key, const Gnome::Conf::Value& value);
 
-    bool rtpRecvAvailable, rtpSendAvailable, zeroconfPublishAvailable, zeroconfDiscoverAvailable, remoteAvailable;
+    void installFiles(const char *a, const char *b);
+
+    bool
+        rtpRecvAvailable,
+        rtpSendAvailable,
+        zeroconfPublishAvailable,
+        zeroconfDiscoverAvailable,
+        zeroconfRaopDiscoverAvailable,
+        remoteAvailable,
+        upnpAvailable;
 };
 
 MainWindow::MainWindow(BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade::Xml>& x) :
     Gtk::Window(cobject), ignoreChanges(true) {
 
-    x->get_widget("titleEventBox", titleEventBox);
     x->get_widget("closeButton", closeButton);
+    x->get_widget("zeroconfDiscoverInstallButton", zeroconfDiscoverInstallButton);
+    x->get_widget("zeroconfRaopDiscoverInstallButton", zeroconfRaopDiscoverInstallButton);
+    x->get_widget("remoteInstallButton", remoteInstallButton);
+    x->get_widget("zeroconfPublishInstallButton", zeroconfPublishInstallButton);
+    x->get_widget("upnpInstallButton", upnpInstallButton);
+    x->get_widget("rtpRecvInstallButton", rtpRecvInstallButton);
+    x->get_widget("rtpSendInstallButton", rtpSendInstallButton);
 
     x->get_widget("remoteAccessCheckButton", remoteAccessCheckButton);
     x->get_widget("zeroconfDiscoverCheckButton", zeroconfDiscoverCheckButton);
+    x->get_widget("zeroconfRaopDiscoverCheckButton", zeroconfRaopDiscoverCheckButton);
     x->get_widget("zeroconfBrowseCheckButton", zeroconfPublishCheckButton);
     x->get_widget("anonymousAuthCheckButton", anonymousAuthCheckButton);
     x->get_widget("rtpReceiveCheckButton", rtpReceiveCheckButton);
     x->get_widget("rtpSendCheckButton", rtpSendCheckButton);
     x->get_widget("rtpLoopbackCheckButton", rtpLoopbackCheckButton);
     x->get_widget("combineCheckButton", combineCheckButton);
+    x->get_widget("upnpMediaServerCheckButton", upnpMediaServerCheckButton);
+    x->get_widget("upnpNullSinkCheckButton", upnpNullSinkCheckButton);
 
     x->get_widget("rtpMikeRadioButton", rtpMikeRadioButton);
     x->get_widget("rtpSpeakerRadioButton", rtpSpeakerRadioButton);
     x->get_widget("rtpNullSinkRadioButton", rtpNullSinkRadioButton);
 
-    Gdk::Color c("white");
-    titleEventBox->modify_bg(Gtk::STATE_NORMAL, c);
-
     checkForModules();
 
     gconf = Gnome::Conf::Client::get_default_client();
@@ -115,6 +160,7 @@ MainWindow::MainWindow(BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade:
     anonymousAuthCheckButton->signal_toggled().connect(sigc::mem_fun(*this, &MainWindow::onChangeRemoteAccess));
 
     zeroconfDiscoverCheckButton->signal_toggled().connect(sigc::mem_fun(*this, &MainWindow::onChangeZeroconfDiscover));
+    zeroconfRaopDiscoverCheckButton->signal_toggled().connect(sigc::mem_fun(*this, &MainWindow::onChangeZeroconfRaopDiscover));
 
     rtpReceiveCheckButton->signal_toggled().connect(sigc::mem_fun(*this, &MainWindow::onChangeRtpReceive));
 
@@ -125,6 +171,17 @@ MainWindow::MainWindow(BaseObjectType* cobject, const Glib::RefPtr<Gnome::Glade:
     rtpNullSinkRadioButton->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::onChangeRtpSend));
 
     combineCheckButton->signal_toggled().connect(sigc::mem_fun(*this, &MainWindow::onChangeCombine));
+
+    upnpMediaServerCheckButton->signal_toggled().connect(sigc::mem_fun(*this, &MainWindow::onChangeUpnp));
+    upnpNullSinkCheckButton->signal_toggled().connect(sigc::mem_fun(*this, &MainWindow::onChangeUpnp));
+
+    zeroconfDiscoverInstallButton->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::onZeroconfDiscoverInstallButtonClicked));
+    zeroconfRaopDiscoverInstallButton->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::onZeroconfRaopDiscoverInstallButtonClicked));
+    remoteInstallButton->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::onRemoteInstallButtonClicked));
+    zeroconfPublishInstallButton->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::onZeroconfPublishInstallButtonClicked));
+    upnpInstallButton->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::upnpInstallButtonClicked));
+    rtpRecvInstallButton->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::rtpRecvInstallButtonClicked));
+    rtpSendInstallButton->signal_clicked().connect(sigc::mem_fun(*this, &MainWindow::rtpSendInstallButtonClicked));
 }
 
 MainWindow* MainWindow::create() {
@@ -147,6 +204,7 @@ void MainWindow::updateSensitive() {
     anonymousAuthCheckButton->set_sensitive(b && remoteAvailable);
 
     zeroconfDiscoverCheckButton->set_sensitive(zeroconfDiscoverAvailable);
+    zeroconfRaopDiscoverCheckButton->set_sensitive(zeroconfRaopDiscoverAvailable);
 
     rtpReceiveCheckButton->set_sensitive(rtpRecvAvailable);
     rtpSendCheckButton->set_sensitive(rtpSendAvailable);
@@ -155,6 +213,44 @@ void MainWindow::updateSensitive() {
     rtpMikeRadioButton->set_sensitive(b && rtpSendAvailable);
     rtpSpeakerRadioButton->set_sensitive(b && rtpSendAvailable);
     rtpNullSinkRadioButton->set_sensitive(b && rtpSendAvailable);
+
+    upnpMediaServerCheckButton->set_sensitive(upnpAvailable);
+    upnpNullSinkCheckButton->set_sensitive(upnpAvailable && upnpMediaServerCheckButton->get_active());
+
+    if (zeroconfDiscoverAvailable)
+        zeroconfDiscoverInstallButton->hide();
+    else
+        zeroconfDiscoverInstallButton->show();
+
+    if (zeroconfRaopDiscoverAvailable)
+        zeroconfRaopDiscoverInstallButton->hide();
+    else
+        zeroconfRaopDiscoverInstallButton->show();
+
+    if (remoteAvailable)
+        remoteInstallButton->hide();
+    else
+        remoteInstallButton->show();
+
+    if (zeroconfPublishAvailable)
+        zeroconfPublishInstallButton->hide();
+    else
+        zeroconfPublishInstallButton->show();
+
+    if (upnpAvailable)
+        upnpInstallButton->hide();
+    else
+        upnpInstallButton->show();
+
+    if (rtpRecvAvailable)
+        rtpRecvInstallButton->hide();
+    else
+        rtpRecvInstallButton->show();
+
+    if (rtpSendAvailable)
+        rtpSendInstallButton->hide();
+    else
+        rtpSendInstallButton->show();
 }
 
 void MainWindow::onChangeRemoteAccess() {
@@ -175,6 +271,15 @@ void MainWindow::onChangeZeroconfDiscover() {
     writeToGConfZeroconfDiscover();
 }
 
+void MainWindow::onChangeZeroconfRaopDiscover() {
+
+    if (ignoreChanges)
+        return;
+
+    updateSensitive();
+    writeToGConfZeroconfRaopDiscover();
+}
+
 void MainWindow::onChangeRtpReceive() {
     if (ignoreChanges)
         return;
@@ -193,14 +298,83 @@ void MainWindow::onChangeRtpSend() {
 }
 
 void MainWindow::onChangeCombine() {
-    Gnome::Conf::ChangeSet changeSet;
-
     if (ignoreChanges)
         return;
 
     writeToGConfCombine();
 }
 
+void MainWindow::onChangeUpnp() {
+
+    if (ignoreChanges)
+        return;
+
+    updateSensitive();
+    writeToGConfUPnP();
+}
+
+void MainWindow::installFiles(const char *a, const char *b = NULL) {
+    DBusGConnection *connection;
+    DBusGProxy *proxy;
+    gboolean ret;
+    GError *error = NULL;
+    const gchar *packages[] = {a, b, NULL};
+
+    connection = dbus_g_bus_get(DBUS_BUS_SESSION, NULL);
+
+    proxy = dbus_g_proxy_new_for_name(connection,
+                                      "org.freedesktop.PackageKit",
+                                      "/org/freedesktop/PackageKit",
+                                      "org.freedesktop.PackageKit.Modify");
+
+    ret = dbus_g_proxy_call(
+            proxy, "InstallProvideFiles", &error,
+            G_TYPE_UINT, GDK_WINDOW_XID(get_window()->gobj()),
+            G_TYPE_STRV, packages,
+            G_TYPE_STRING, "show-confirm-search,hide-finished",
+            G_TYPE_INVALID, G_TYPE_INVALID);
+
+    if (!ret) {
+        g_warning("Installation failed: %s", error->message);
+        g_error_free(error);
+    }
+
+    g_object_unref(proxy);
+    dbus_g_connection_unref(connection);
+
+    checkForModules();
+    updateSensitive();
+}
+
+void MainWindow::onZeroconfDiscoverInstallButtonClicked() {
+    installFiles(MODULESDIR "module-zeroconf-discover" SHREXT);
+}
+
+void MainWindow::onZeroconfRaopDiscoverInstallButtonClicked() {
+    installFiles(MODULESDIR "module-raop-discover" SHREXT);
+}
+
+void MainWindow::onRemoteInstallButtonClicked() {
+    installFiles(MODULESDIR "module-esound-protocol-tcp" SHREXT,
+                 MODULESDIR "module-native-protocol-tcp" SHREXT);
+}
+
+void MainWindow::onZeroconfPublishInstallButtonClicked() {
+    installFiles(MODULESDIR "module-zeroconf-publish" SHREXT);
+}
+
+void MainWindow::upnpInstallButtonClicked() {
+    installFiles("/usr/bin/rygel", MODULESDIR "module-rygel-media-server" SHREXT);
+}
+
+void MainWindow::rtpRecvInstallButtonClicked() {
+    installFiles(MODULESDIR "module-rtp-recv" SHREXT);
+}
+
+void MainWindow::rtpSendInstallButtonClicked() {
+    installFiles(MODULESDIR "module-rtp-send" SHREXT);
+}
+
 void MainWindow::writeToGConfCombine() {
     Gnome::Conf::ChangeSet changeSet;
     changeSet.set(PA_GCONF_PATH_MODULES"/combine/locked", true);
@@ -286,6 +460,28 @@ void MainWindow::writeToGConfZeroconfDiscover() {
     gconf->suggest_sync();
 }
 
+void MainWindow::writeToGConfZeroconfRaopDiscover() {
+    Gnome::Conf::ChangeSet changeSet;
+
+    changeSet.set(PA_GCONF_PATH_MODULES"/raop-discover/locked", true);
+    gconf->change_set_commit(changeSet, true);
+
+    if (zeroconfRaopDiscoverCheckButton->get_active()) {
+        changeSet.set(PA_GCONF_PATH_MODULES"/raop-discover/name0", Glib::ustring("module-raop-discover"));
+        changeSet.set(PA_GCONF_PATH_MODULES"/raop-discover/args0", Glib::ustring(""));
+
+        changeSet.set(PA_GCONF_PATH_MODULES"/raop-discover/enabled", true);
+    } else
+        changeSet.set(PA_GCONF_PATH_MODULES"/raop-discover/enabled", false);
+
+    gconf->change_set_commit(changeSet, true);
+
+    changeSet.set(PA_GCONF_PATH_MODULES"/raop-discover/locked", false);
+    gconf->change_set_commit(changeSet, true);
+
+    gconf->suggest_sync();
+}
+
 void MainWindow::writeToGConfRtpReceive() {
     Gnome::Conf::ChangeSet changeSet;
 
@@ -324,7 +520,11 @@ void MainWindow::writeToGConfRtpSend() {
     if (rtpSendCheckButton->get_active()) {
         if (!mikeEnabled && !speakerEnabled) {
             changeSet.set(PA_GCONF_PATH_MODULES"/rtp-send/name0", Glib::ustring("module-null-sink"));
-            changeSet.set(PA_GCONF_PATH_MODULES"/rtp-send/args0", Glib::ustring("sink_name=rtp format=s16be channels=2 rate=44100 description=\"RTP Multicast Sink\""));
+            changeSet.set(PA_GCONF_PATH_MODULES"/rtp-send/args0", Glib::ustring("sink_name=rtp "
+                                                                                "format=s16be "
+                                                                                "channels=2 "
+                                                                                "rate=44100 "
+                                                                                "sink_properties=\"device.description='RTP Multicast' device.bus='network' device.icon_name='network-server'\""));
 
             changeSet.set(PA_GCONF_PATH_MODULES"/rtp-send/name1", Glib::ustring("module-rtp-send"));
             changeSet.set(PA_GCONF_PATH_MODULES"/rtp-send/args1", Glib::ustring(loopbackEnabled ? "source=rtp.monitor loop=1" : "source=rtp.monitor loop=0"));
@@ -354,6 +554,42 @@ void MainWindow::writeToGConfRtpSend() {
     gconf->suggest_sync();
 }
 
+void MainWindow::writeToGConfUPnP() {
+    Gnome::Conf::ChangeSet changeSet;
+
+    changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/locked", true);
+    gconf->change_set_commit(changeSet, true);
+
+    if (upnpMediaServerCheckButton->get_active()) {
+        changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/name0", Glib::ustring("module-rygel-media-server"));
+        changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/args0", Glib::ustring(""));
+
+        if (upnpNullSinkCheckButton->get_active()) {
+            changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/name1", Glib::ustring("module-null-sink"));
+            changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/args1", Glib::ustring("sink_name=upnp "
+                                                                                         "format=s16be "
+                                                                                         "channels=2 "
+                                                                                         "rate=44100 "
+                                                                                         "sink_properties=\"device.description='DLNA/UPnP Streaming' device.bus='network' device.icon_name='network-server'\""));
+            changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/null-sink-enabled", true);
+        } else {
+            changeSet.unset(PA_GCONF_PATH_MODULES"/upnp-media-server/name1");
+            changeSet.unset(PA_GCONF_PATH_MODULES"/upnp-media-server/args1");
+            changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/null-sink-enabled", false);
+        }
+
+        changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/enabled", true);
+    }  else
+        changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/enabled", false);
+
+    gconf->change_set_commit(changeSet, true);
+
+    changeSet.set(PA_GCONF_PATH_MODULES"/upnp-media-server/locked", false);
+    gconf->change_set_commit(changeSet, true);
+
+    gconf->suggest_sync();
+}
+
 void MainWindow::onGConfChange(const Glib::ustring&, const Gnome::Conf::Value&) {
     readFromGConf();
 }
@@ -368,6 +604,7 @@ void MainWindow::readFromGConf() {
     anonymousAuthCheckButton->set_active(gconf->get_bool(PA_GCONF_PATH_MODULES"/remote-access/anonymous_enabled"));
 
     zeroconfDiscoverCheckButton->set_active(gconf->get_bool(PA_GCONF_PATH_MODULES"/zeroconf-discover/enabled"));
+    zeroconfRaopDiscoverCheckButton->set_active(gconf->get_bool(PA_GCONF_PATH_MODULES"/raop-discover/enabled"));
 
     rtpReceiveCheckButton->set_active(gconf->get_bool(PA_GCONF_PATH_MODULES"/rtp-recv/enabled"));
 
@@ -384,6 +621,9 @@ void MainWindow::readFromGConf() {
 
     combineCheckButton->set_active(gconf->get_bool(PA_GCONF_PATH_MODULES"/combine/enabled"));
 
+    upnpMediaServerCheckButton->set_active(gconf->get_bool(PA_GCONF_PATH_MODULES"/upnp-media-server/enabled"));
+    upnpNullSinkCheckButton->set_active(gconf->get_bool(PA_GCONF_PATH_MODULES"/upnp-media-server/null-sink-enabled"));
+
     ignoreChanges = FALSE;
 
     updateSensitive();
@@ -398,8 +638,14 @@ void MainWindow::checkForModules() {
     zeroconfPublishAvailable = access(MODULESDIR "module-zeroconf-publish" SHREXT, F_OK) == 0;
     zeroconfDiscoverAvailable = access(MODULESDIR "module-zeroconf-discover" SHREXT, F_OK) == 0;
 
+    zeroconfRaopDiscoverAvailable = access(MODULESDIR "module-raop-discover" SHREXT, F_OK) == 0;
+
     rtpRecvAvailable = access(MODULESDIR "module-rtp-recv" SHREXT, F_OK) == 0;
     rtpSendAvailable = access(MODULESDIR "module-rtp-send" SHREXT, F_OK) == 0;
+
+    upnpAvailable =
+        access(MODULESDIR "module-rygel-media-server" SHREXT, F_OK) == 0 &&
+        g_find_program_in_path("rygel");
 }
 
 int main(int argc, char *argv[]) {
